//*************************************************************************
//* ---------------------
//* CDCClusterMaker Class
//* ---------------------
//* (Description)
//*	This set of classes makes FADC clusters.
//* (Requires)
//*	TObjArray:
//*	JSFSteer:
//*	JSFModules:
//*	CDCRawAData:
//* (Provides)
//*	CDCClusterMaker		: steering class
//*	CDCClusterBuf		: buffer to store clusters
//*	CDCCluster		: cluster data class
//*	CDCFADCCalibration	: FADC calibration
//* (Update Record)
//*	1999/12/04  K.Fujii	Original version.
//*	2000/06/07  K.Fujii	Added software threshold.
//*	2000/07/15  K.Fujii	Added fGlobalClusters to CDCClusterBuf and
//*				fParent, fSplitThresh, and SplitInto() to
//*				CDCCluster.
//*	2000/07/19  K.Fujii	Minor change in SplitInto() method.
//*	2000/09/30  K.Fujii	Fixed a bug in SplitInto() method:
//*				No notyet check has been done before
//*				updating vly.
//*	2000/11/10  K.Hoshina	Restrict No. of time slots for PH sum.
//*************************************************************************

#include "CDCClusterMaker.h"
#include <fstream.h>

Int_t CDCCluster::fNpres  = 100; 
Int_t CDCCluster::fNposts = 100; 
Int_t CDCCluster::fMaxNslots = 999999; 
Int_t CDCCluster::fSortBy =  0; // (0,1) = (time,charge)
Double_t CDCCluster::fSoftThresh  = 16.;
Double_t CDCCluster::fSplitThresh = 16.;

//_____________________________________________________________________
// ---------------------
// FADC Calibration Data
// ---------------------

ClassImp(CDCFADCCalibration)

CDCFADCCalibration::CDCFADCCalibration(Double_t *ped, Double_t *gain, 
  		     		       Double_t *tped, Double_t *tscale, Int_t last)
{
  	CalibDatum *datum = new CalibDatum(last,ped,gain,tped,tscale);
  	Add(datum);
  	fCurr = datum;
}

CDCFADCCalibration::CDCFADCCalibration(Char_t * calibfile) 
{
	cout << " Going to open a calib file " << calibfile << endl;
	ifstream in;
	in.open(calibfile,ios::in);
	cout << " Opened " << calibfile << endl;
        Int_t last = 99999999;
        Double_t *ped    = new Double_t [CDCRawADCDatum::GetNchMax()];
        Double_t *gain   = new Double_t [CDCRawADCDatum::GetNchMax()];
        Double_t *tped   = new Double_t [CDCRawADCDatum::GetNchMax()];
        Double_t *tscale = new Double_t [CDCRawADCDatum::GetNchMax()];
	while ((in >> last 
		   >> ped   [0] >> ped   [1] >> ped   [2]  >> ped   [3]
        	   >> ped   [4] >> ped   [5] >> ped   [6]  >> ped   [7]
        	   >> ped   [8] >> ped   [9] >> ped   [10] >> ped   [11]) &&
	       (in >> gain  [0] >> gain  [1] >> gain  [2]  >> gain  [3]
        	   >> gain  [4] >> gain  [5] >> gain  [6]  >> gain  [7]
        	   >> gain  [8] >> gain  [9] >> gain  [10] >> gain  [11]) &&
	       (in >> tped  [0] >> tped  [1] >> tped  [2]  >> tped  [3]
        	   >> tped  [4] >> tped  [5] >> tped  [6]  >> tped  [7]
        	   >> tped  [8] >> tped  [9] >> tped  [10] >> tped  [11]) &&
	       (in >> tscale[0] >> tscale[1] >> tscale[2]  >> tscale[3]
        	   >> tscale[4] >> tscale[5] >> tscale[6]  >> tscale[7]
        	   >> tscale[8] >> tscale[9] >> tscale[10] >> tscale[11])) {
        	CalibDatum *datum = new CalibDatum(last,ped,gain,tped,tscale);
		Add(datum);
        }
	in.close();
	delete [] ped;
	delete [] gain;
	delete [] tped;
	delete [] tscale;
}

CDCFADCCalibration::~CDCFADCCalibration() { Delete(); }
  
Double_t CDCFADCCalibration::GetPedestal    (Int_t n, Int_t a) 
  		{ return fCurr->GetPedestal(n,a);    }
Double_t CDCFADCCalibration::GetGain        (Int_t n, Int_t a)
  		{ return fCurr->GetGain(n,a);        }
Double_t CDCFADCCalibration::GetTimePedestal(Int_t n, Int_t a)
  		{ return fCurr->GetTimePedestal(n,a);}
Double_t CDCFADCCalibration::GetTimeScale   (Int_t n, Int_t a)
  		{ return fCurr->GetTimeScale(n,a);   }

void CDCFADCCalibration::Update()
{
  	Int_t run = gJSF->GetRunNumber();
  	TIter next(this);
  	CalibDatum *datum;
  	while ((datum = (CalibDatum *)next())) {
  		fCurr = datum;
  		if (run <= fCurr->GetLastRun()) break;
  	}	
}


//_____________________________________________________________________
// -------------
// Cluster Datum
// -------------

ClassImp(CDCCluster)

CDCCluster::CDCCluster(CDCFADCCalibration *calib, Bool_t varped) : TObjArray(1), 
	fDone(kFALSE), fUseVarPed(varped), 
	fStationNo(0), fChannelNo(0),
	fStartTimeSlot(0), fLastTimeSlot(0),
	fSum(0.), fPeak(0.), fPedestal(0.),
	fStartTime(0.), fWidth(0.),
	fNpeds((fNpres - 20 > 0 ? fNpres - 20 : 10)), 
	fCalib(calib), fParent(0) {}

CDCCluster::CDCCluster(CDCCluster & cl) : TObjArray(cl),
	fDone(cl.fDone), fUseVarPed(cl.fUseVarPed),
	fStationNo(cl.fStationNo), fChannelNo(cl.fChannelNo),
	fStartTimeSlot(cl.fStartTimeSlot), fLastTimeSlot(cl.fLastTimeSlot),
	fSum(cl.fSum), fPeak(cl.fPeak), fPedestal(cl.fPedestal),
	fStartTime(cl.fStartTime), fWidth(cl.fWidth),
	fNpeds(cl.fNpeds), 
	fCalib(cl.fCalib), fParent(0) {}

CDCCluster::~CDCCluster() {}

void CDCCluster::SetNpeds(Int_t n)
{
  	if (fNpeds != n) {
  		fNpeds = n;
  		fDone  = kFALSE;
  	}
  	if (!fDone) Update();
}

Double_t CDCCluster::GetPedestal()
{
  	if (!fUseVarPed) {
   		return fCalib->GetPedestal(fStationNo,fChannelNo);
  	} else {
  	if (!fDone) Update(); 
  		return fPedestal;
  	}
}

void CDCCluster::UseVarPed()
{
	if (!fUseVarPed) {
		fUseVarPed = kTRUE;
		fDone = kFALSE;
	}
	if (!fDone) Update();
}

void CDCCluster::UseConstPed()
{
	if (fUseVarPed) {
		fUseVarPed = kFALSE;
		fDone = kFALSE;
	}
	if (!fDone) Update();
}


Bool_t CDCCluster::Append(CDCRawADCDatum *adc)
{
	if (GetEntries() == 0) {
		fStationNo     = adc->GetStationNo();
		fChannelNo     = adc->GetChannelNo();
		fStartTimeSlot = adc->GetTime();
		fLastTimeSlot  = fStartTimeSlot;
		Add(adc);
		return kTRUE;
	} else {
		if (adc->GetStationNo() == fStationNo &&
		    adc->GetChannelNo() == fChannelNo &&
		    adc->GetTime() == fLastTimeSlot+1) {
			fLastTimeSlot++;
			Add(adc);
			return kTRUE;
		} else {
			return kFALSE;
		}
	}
			
}

void CDCCluster::SplitInto(TObjArray &list)
{
	if (fParent) return;
	
	CDCCluster *cur = new CDCCluster(fCalib,fUseVarPed);
	cur->fParent = this;
#if 0
	Int_t nslots = GetEntries();
	Int_t islot;
	for (islot = fNpres; islot < nslots - fNposts; islot++) {
		cur->Append((CDCRawADCDatum *)At(islot));
	}
	
	if (cur->GetEntries() > 0) list.Add(cur);
	else			   delete cur;
#else
	Double_t ped  = fCalib->GetPedestal(fStationNo,fChannelNo);
	if (fUseVarPed) ped = fPedestal;

	Double_t vly  = 0.;
	Double_t pek  = -99999.;

	Bool_t   notyet = kTRUE;	// peak not yet found
	Int_t    i1st   = fNpres;
	Int_t    ilst   = fNpres;
	Int_t    ipek   = fNpres;
	Int_t    in1st  = fNpres;

	Int_t nslots = GetEntries();
	Int_t islot;
	for (islot = fNpres; islot < nslots - fNposts; islot++) {
		CDCRawADCDatum *adclp = (CDCRawADCDatum *)At(islot-1);
		CDCRawADCDatum *adccp = (CDCRawADCDatum *)At(islot);
		CDCRawADCDatum *adcrp = (CDCRawADCDatum *)At(islot+1);
		
		Double_t adcl = ped - adclp->GetADC();
		Double_t adcc = ped - adccp->GetADC();
		Double_t adcr = ped - adcrp->GetADC();
			
		if (notyet && adcc < pek - fSplitThresh) { // peak found
			notyet = kFALSE;
			vly    = pek;
		}
		
		if (!notyet && adcc > vly + fSplitThresh) { // valley deep enough
			Int_t i;			    // save the previous
			for (i = i1st; i <= ilst; i++) {    // peak
				cur->Append((CDCRawADCDatum *)At(i));
			}
			if (cur->GetEntries() > 0) {
				list.Add(cur);
				cur = new CDCCluster(fCalib,fUseVarPed);
				cur->fParent = this;
			}
			notyet = kTRUE;
			i1st   = TMath::Max(in1st,ilst + 1);
			if (ipek < i1st) {
				pek  = vly;
				ipek = ilst;
			}
		}


		if (adcc > pek && adcc >= adcl && adcc >= adcr) {
			pek  = adcc;		// peak candidate
			ipek = islot;
		}
		
		if (adcc < vly && adcc <= adcl && adcc <= adcr) {
			vly  = adcc;		// valley candidate
			if (notyet) { 
				i1st = islot;  	// 1st slot candidate
			} else {
				ilst = islot;	// last slot chandidate
			}
		}

		if (adcc < fSoftThresh && adcc <= adcl && adcc <= adcr) {
			if (notyet) i1st  = islot;
			else        in1st = islot;
		}
		
		if (!notyet && islot == (nslots - fNposts -1)) ilst = islot;
	}
	if (ilst > i1st) {	// save the left over
		Int_t i;
		for (i = i1st; i <= ilst; i++) {
			cur->Append((CDCRawADCDatum *)At(i));
		}
		list.Add(cur);
	} else {
		if (cur) delete cur;
	}
#endif
}


void CDCCluster::Update()
{
	if (fDone) return;
	fDone = kTRUE;
	Double_t ped    = fCalib->GetPedestal(fStationNo,fChannelNo);
	Double_t gain   = fCalib->GetGain(fStationNo,fChannelNo);
	Double_t tped   = fCalib->GetTimePedestal(fStationNo,fChannelNo);
	Double_t tscale = fCalib->GetTimeScale(fStationNo,fChannelNo);

	if (!fParent) {
		TIter next(this);
		Int_t nslots = GetEntries();
		if (nslots <= fNpres+fNposts) {
			fPedestal  = ped;
			fPeak      = 0.;
			fSum       = 0.;
			fStartTime = (fStartTimeSlot - tped)*tscale;
			fWidth     = (nslots - fNpres - fNposts)*tscale;
			cerr << "???? CDCCluster::Update() ???? " 
			     << "cluster too narrow: "
			     << "Nslots = " << nslots << endl;
			DebugPrint("Detailed");
			return;
		}
	
		fPedestal = 0;
		CDCRawADCDatum *adcp;
		Int_t islot;
		for (islot = 0; islot < fNpres; islot++) {
			adcp = (CDCRawADCDatum *)next();
			if (islot < fNpeds) fPedestal += adcp->GetADC();
		}
		fPedestal /= fNpeds;
	
		if (fUseVarPed) ped = fPedestal;
	
		fSum       = 0;
		fPeak      = -99999.;
		fStartTime = fStartTimeSlot;
		for (; islot < nslots - fNposts; islot++) {
			adcp = (CDCRawADCDatum *)next();
			Double_t adc = ped - adcp->GetADC();
			fPeak  = TMath::Max(fPeak,adc);
			if (fStartTime == fStartTimeSlot && adc > fSoftThresh) {
				fStartTime += islot - fNpres;
			}
			fSum  += adc;
		}
		fSum *= gain;
	
		fStartTime = (fStartTime - tped)*tscale;
		fWidth     = (nslots - fNpres - fNposts)*tscale;
	} else {
		if (!fParent->fDone) fParent->Update();
		fPedestal = fParent->fPedestal;
		if (fUseVarPed) ped = fPedestal;

		TIter next(this);
		Int_t nslots = TMath::Min(GetEntries(), fMaxNslots);
		
		fSum       = 0;
		fPeak      = -99999.;
		fStartTime = fStartTimeSlot;
		Int_t islot;
		CDCRawADCDatum *adcp;
		for (islot = 0; islot < nslots; islot++) {
			adcp = (CDCRawADCDatum *)next();
			Double_t adc = ped - adcp->GetADC();
			fPeak  = TMath::Max(fPeak,adc);
			if (fStartTime == fStartTimeSlot && adc > fSoftThresh) {
				fStartTime += islot - fNpres;
			}
			fSum  += adc;
		}
		fSum *= gain;
	
		fStartTime = (fStartTime - tped)*tscale;
		fWidth     = nslots*tscale;
	}
}

Bool_t CDCCluster::IsSortable() const { return kTRUE; }

Int_t CDCCluster::Compare(TObject *obj)
{
	if (!fDone) Update();
	if (!((CDCCluster *)obj)->fDone) ((CDCCluster *)obj)->Update();
	switch (fSortBy) {
	    case 1: // sort by charge:  bigger one first:-)
   		if (fSum > ((CDCCluster *)obj)->fSum) return -1;
   		else if (fSum < ((CDCCluster *)obj)->fSum) return 1;
  		else return 0;
	    case 2: // sort by pulse height:  bigger one first:-)
   		if (fPeak > ((CDCCluster *)obj)->fPeak) return -1;
   		else if (fPeak < ((CDCCluster *)obj)->fPeak) return 1;
  		else return 0;
	    default: // sort by time: faster one first:-)
   		if (fStartTime < ((CDCCluster *)obj)->fStartTime) return -1;
   		else if (fStartTime > ((CDCCluster *)obj)->fStartTime) return 1;
  		else return 0;
  	}
}

void CDCCluster::SetSortBy(Char_t *opt)
{
	switch (opt[0]) {
		case 'C':	// sort by "Charge"
			fSortBy = 1;
			break;
		case 'P': 	// sort by "PulseHeight"
			fSortBy = 2;
			break;
		default: 	// sort by "Time"
			fSortBy = 0;
	}		
}

void CDCCluster::DebugPrint(const Char_t *opt)
{
	cerr.setf(ios::right);
	cerr.setf(ios::fixed);
	cerr	<< " (n,a)=(" << setw(2) << GetStationNo() << ","
			     << setw(1) << GetChannelNo() << ")"
		<< " parent = " << setw(5) << (void *)fParent
   		<< " t = " << setw(6) << setprecision(1) << GetStartTime()
   		<< " w = " << setw(6) << setprecision(1) << GetWidth()
   		<< " Sum = " << setw(7) << setprecision(1) << GetSum()
    		<< " Pek = " << setw(5) << setprecision(1) << GetPeak()
    		<< " Ped = " << setw(5) << setprecision(1) << GetPedestal()
    		<< endl;
	cerr.unsetf(ios::right);
	cerr.unsetf(ios::fixed);

	if (strstr(opt,"Detailed")) {
		TIter next(this);
		CDCRawADCDatum *adc;
		while ((adc = (CDCRawADCDatum *)next())) {
			adc->DebugPrint();
		}
    	}
}

//_____________________________________________________________________
// --------------
// Cluster Buffer
// --------------

ClassImp(CDCClusterBuf)

CDCClusterBuf::CDCClusterBuf(CDCFADCCalibration *calib,
			     const char *name,const char *title, 
			     CDCClusterMaker *module) :
	JSFEventBuf(name, title, (JSFModule*)module),
	fDone(kFALSE), fUseVarPed(kTRUE), fNADCData(0), fADCData(0), 
	fGlobalClusters(1), fClusters(1), 
	fCalib(calib) {}

CDCClusterBuf::CDCClusterBuf(CDCClusterMaker *module,
			     CDCFADCCalibration *calib,
			     const char *name, const char *title) :
	JSFEventBuf(name, title, (JSFModule*)module),
	fDone(kFALSE), fUseVarPed(kTRUE), fNADCData(0), fADCData(0), 
	fGlobalClusters(1), fClusters(1), 
	fCalib(calib) {}

CDCClusterBuf::~CDCClusterBuf()
{
	fGlobalClusters.Delete();
	fClusters.Delete();
}

void CDCClusterBuf::SetADCData(Len_t nadcdata, TClonesArray *adcdata)
{
	fDone     = kFALSE;
	fNADCData = nadcdata;
	fADCData  = adcdata;
	if (fGlobalClusters.GetEntries() > 0) 	fGlobalClusters.Delete();
	if (fClusters.GetEntries() > 0) 	fClusters.Delete();
}

void CDCClusterBuf::SetCalibrationData(CDCFADCCalibration *calib)
{
	fDone  = kFALSE;
	fCalib = calib;
}

Bool_t CDCClusterBuf::IsInitialized()  const
{
	if (fNADCData > 0) return kTRUE;
	else return kFALSE;
}

Int_t CDCClusterBuf::GetNclusters()
{
	if (!fDone) FindClusters(); 
	return fClusters.GetEntries();
}

TObjArray * CDCClusterBuf::GetClusters()
{
	if (!fDone) FindClusters();
	return &fClusters;
}

void CDCClusterBuf::UseVarPed()
{
	if (!IsInitialized()) {
		fUseVarPed = kTRUE;
	} else {
		if (!fUseVarPed) {
			fDone      = kFALSE;
			fClusters.Delete();
			fUseVarPed = kTRUE;
			TIter next(&fGlobalClusters);
			CDCCluster *clp;
			while ((clp = (CDCCluster *)next())) {
				clp->UseVarPed();
			}
		}
	}
}

void CDCClusterBuf::UseConstPed()
{
	if (!IsInitialized()) {
		fUseVarPed = kFALSE;
	} else {
		if (fUseVarPed) {
			fDone      = kFALSE;
			fClusters.Delete();
			fUseVarPed = kFALSE;
			TIter next(&fClusters);
			CDCCluster *clp;
			while ((clp = (CDCCluster *)next())) {
				clp->UseConstPed();
			}
		}
	}
}

void CDCClusterBuf::FindClusters()
{
	if (!IsInitialized()) {
		cerr << "ERROR  CDCClusterBuf::FindCluster" << endl
		     << "...... ADC data not yet initialized." << endl;
		return;
	} else if (fDone || fNADCData == 0) {
		return;
	}
	
	fDone = kTRUE;
	if (fGlobalClusters.GetEntries() == 0) {
		CDCCluster *cur = new CDCCluster(fCalib,fUseVarPed);
		fGlobalClusters.Add(cur);
	
		TIter next(fADCData);
		CDCRawADCDatum *adcp;
		while ( (adcp = (CDCRawADCDatum *)next()) ) {
			if (adcp->GetChannelNo() < 0 ||
		    	adcp->GetChannelNo() >= CDCRawADCDatum::GetNchMax()) {
		    		// adcp->DebugPrint();
		    		continue; // skip hodoscopes
			}
			if (! cur->Append(adcp)) {
				cur = new CDCCluster(fCalib,fUseVarPed);
				fGlobalClusters.Add(cur);
				cur->Append(adcp);
			}
		}
	}
	
	if (fGlobalClusters.GetEntries() == 0) return;

	TIter next(&fGlobalClusters);
	CDCCluster *clp;
	while ( (clp = (CDCCluster *)next()) ) {
		clp->SplitInto(fClusters);	// split multi-peaks
	}
}

void CDCClusterBuf::DebugPrint(const Char_t *opt)
{
	cerr	<< fClusters.GetEntries()
		<< " clusters in this event." << endl;
	TIter next(&fClusters);
	CDCCluster *clp;
	Int_t n = 0;
	while ((clp = (CDCCluster *)next())) {
		cerr << setw(3) << ++n << " ";
		clp->DebugPrint(opt);
	}
}


//_____________________________________________________________________
// -------------
// Cluster Maker
// -------------

ClassImp(CDCClusterMaker)

CDCClusterMaker::CDCClusterMaker(CDCFADCCalibration *calib,
				 const char *name, const char *title)
       : JSFModule(name,title),
         fToDeleteCalib(calib ? kFALSE : kTRUE),
         fCalib(calib) {}

CDCClusterMaker::~CDCClusterMaker()
{
	if (fToDeleteCalib && fCalib) delete fCalib;
}

void CDCClusterMaker::SetCalibrationData(CDCFADCCalibration *calib)
{
	fCalib = calib;
	fToDeleteCalib = kFALSE;
	((CDCClusterBuf *)EventBuf())->SetCalibrationData(calib);
}

void CDCClusterMaker::UseVarPed()
{
	fUseVarPed = kTRUE;
	if (EventBuf()) ((CDCClusterBuf *)EventBuf())->UseVarPed();
}

void CDCClusterMaker::UseConstPed()
{
	fUseVarPed = kFALSE;
	if (EventBuf()) ((CDCClusterBuf *)EventBuf())->UseConstPed();
}

Bool_t CDCClusterMaker::Initialize()
{
	if (!fCalib) {
		fCalib = new CDCFADCCalibration();
		fToDeleteCalib = kTRUE;
	}
	fEventBuf = new CDCClusterBuf(this,fCalib);
	if (fUseVarPed) ((CDCClusterBuf *)fEventBuf)->UseVarPed();
	else		((CDCClusterBuf *)fEventBuf)->UseConstPed();
	
	return kTRUE;
}

Bool_t CDCClusterMaker::BeginRun(Int_t run)
{
	fCalib->Update();
	
	return kTRUE;
}

Bool_t CDCClusterMaker::Process(Int_t ev)
{
	CDCRawData    *raw    = (CDCRawData*)gJSF->FindModule("CDCRawData");
	CDCRawDataBuf *evt    = (CDCRawDataBuf *)raw->EventBuf();
	Int_t         nadc    = evt->GetNADCData();
	TClonesArray *adcdata = evt->GetADCData();

	CDCClusterBuf *clbufp = (CDCClusterBuf *)EventBuf();
	clbufp->SetADCData(nadc,adcdata);	
	clbufp->FindClusters();

	return kTRUE;
}

